=begin

MM5 interface for NumRu::GPhys
by Shigenori OTSUKA

last modified: 2008.9.3

=end

require "numru/gphys"
require "numru/mm5_io_raw"
include NumRu

module NumRu
  class MM5Var
    def initialize(aMM5IO,var,opts={})
      raise ArgumentError,"Not a MM5IO" if ! aMM5IO.is_a?(MM5IO)
      raise ArgumentError,"Not a String" if ! var.is_a?(String)
      @mm5io = aMM5IO
      @var = var
      @axis = (opts[:axis] ? opts[:axis] : false)
    end
    
    def _read(t)
      return @mm5io.read(@var,t)
    end
    
    def val
      if @axis
        return _read(0)
      else
        result = []
        for i in 0...(@mm5io.ntime)
          result.push(_read(i))
        end
        return NArray.to_na(result)
      end
    end
    
    def [](*index)
      if @axis
        return val[*index]
      else
        case index[-1]
        when true,false,0..-1,0...(@mm5io.ntime)
          return val[*index]
        when Numeric
          return _read(index[-1])[*(index[0..-2])]
        when Array, NArray
          result = []
          for i in index[-1]
            result.push(_read(i)[*(index[0..-2])])
          end
          return NArray.to_na(result)
        end
      end
    end
    
    def shape
      if @axis
        @mm5io.get_shape(@var,0)
      else
        @mm5io.get_shape(@var,0).push(@mm5io.ntime)
      end
    end
    
    def total
      len = 1
      @mm5io.get_shape(@var,0).each{|i| len *= i}
      if @axis
        return len
      else
        return len *= @mm5io.ntime
      end
    end
    alias length total
    
    def rank
      rank = @mm5io.get_shape(@var,0).length
      rank += 1 if !@axis
      return rank
    end

    def typecode
      NArray::SFLOAT
    end
  end
  ########################

  class VArrayMM5 < VArray
    def initialize(aMM5IO,var,opts={})
      raise ArgumentError,"Not a MM5IO" if ! aMM5IO.is_a?(MM5IO)
      raise ArgumentError,"Not a String" if ! var.is_a?(String)
      @name = var
      @mapping = nil
      @varray = nil
      @ary = MM5Var.new(aMM5IO,var,opts)
      @attr = nil
    end

    def val
      @ary.val
    end

    def inspect
      "<'#{name}' in MM5 sfloat#{ary.shape.inspect}>"
    end
  end
  ###########

  class GPhys
    module MM5_IO

      module_function

      def open(files,var)
        case files
        when MM5IO
          files = [files]
          vary = VArrayMM5.new(files[0],var)
        when String
          files = [MM5IO.open(files)] # is there any side-effect?
          vary = VArrayMM5.new(files[0],var)
        when Array
          files = files.flatten.uniq.collect{|file|
            case file
            when MM5IO
              file
            when String
              MM5IO.open(file)
            else
              raise "only Array of MM5IO or String are acceptable"
            end
          }
          # AUTO-SORT FILES
          files.sort!{|a,b| a.xtimes[0] - b.xtimes[0]}
          # MAKE VArray
          varys = []
          files.each{|file|
            varys.push(VArrayMM5.new(file,var))
          }
          # CHECK CONSISTENCY
          shape = varys[0].shape
          varys.each{|v|
            tmp_shape = v.shape
            shape[0..-2].each_with_index{|s,i|
              raise "shapes are inconsystent" if tmp_shape[i] != s
            }
          }
          # ASSUME CONCATENATION ALONG TIME AXIS
          varys = NArray.to_na(varys).newdim(*([0]*(shape.length-1)))
          vary = VArrayComposite.new(varys)
        else
          raise "only String, MM5IO, and Array are acceptable"
        end
        staggering = files[0].staggering(var,0).strip!
        ordering = files[0].ordering(var,0).strip!
        xtimes = []
        files.each{|file| xtimes.push(file.xtimes)}
      
        shape = vary.shape
        rank = shape.length

        case ordering
        when "YX"
          axisnames = ["x","y","t"]
        when "YXS"
          axisnames = ["x","y","z","t"]
        when "YXW"
          axisnames = ["x","y","z","t"]
        else
          raise "ordering #{ordering} not impremented yet"
        end

        if axisnames.include?("x") && axisnames.include?("y")
          x = NArray.sfloat(shape[axisnames.index("x")]).indgen!
          y = NArray.sfloat(shape[axisnames.index("y")]).indgen!
          vx = VArray.new( x ).rename("x")
          vy = VArray.new( y ).rename("y")
          xax = Axis.new().set_pos(vx)
          yax = Axis.new().set_pos(vy)
          xygrid = Grid.new(xax,yax)
          if staggering == "D"
            lon_c = files[0].read("LONGICRS",0)
            lat_c = files[0].read("LATITCRS",0)
            lon_d = NArray.sfloat(shape[axisnames.index("x")],
                                  shape[axisnames.index("y")])
            lat_d = NArray.sfloat(shape[axisnames.index("x")],
                                  shape[axisnames.index("y")])
            lon_d[1..-2,1..-2] = (lon_c[0..-2,0..-2]\
                                  +lon_c[1..-1,0..-2]\
                                  +lon_c[0..-2,1..-1]\
                                  +lon_c[1..-1,1..-1])*0.25
            lon_d[0,true] = lon_d[1,true]*2 - lon_d[2,true]
            lon_d[true,0] = lon_d[true,1]*2 - lon_d[true,2]
            lon_d[-1,true] = lon_d[-2,true]*2 - lon_d[-3,true]
            lon_d[true,-1] = lon_d[true,-2]*2 - lon_d[true,-3]
            lat_d[1..-2,1..-2] = (lat_c[0..-2,0..-2]\
                                  +lat_c[1..-1,0..-2]\
                                  +lat_c[0..-2,1..-1]\
                                  +lat_c[1..-1,1..-1])*0.25
            lat_d[0,true] = lat_d[1,true]*2 - lat_d[2,true]
            lat_d[true,0] = lat_d[true,1]*2 - lat_d[true,2]
            lat_d[-1,true] = lat_d[-2,true]*2 - lat_d[-3,true]
            lat_d[true,-1] = lat_d[true,-2]*2 - lat_d[true,-3]
            vlon = VArray.new( lon_d ).rename("longitude")
            vlat = VArray.new( lat_d ).rename("latitude")
          else
            vlon = VArrayMM5.new(files[0],"LONGICRS",{:axis=>true})
            vlat = VArrayMM5.new(files[0],"LATITCRS",{:axis=>true})
          end
          glon = GPhys.new(xygrid, vlon) 
          glat = GPhys.new(xygrid, vlat) 
        end
        if axisnames.include?("z")
          z = NArray.sfloat(shape[axisnames.index("z")]).indgen!
          vz = VArray.new( z ).rename("z")
          zax = Axis.new().set_pos(vz)
          if ordering.index("W")
            sigma = files[0].read("SIGMAH",0)
            tmp_sigma = NArray.sfloat(shape[axisnames.index("z")])
            tmp_sigma[1..-2] = (sigma[0..-2]+sigma[1..-1])*0.5
            tmp_sigma[0] = 0
            tmp_sigma[-1] = 1
            vsigma = VArray.new( tmp_sigma ).rename("sigma")
          else
            vsigma = VArrayMM5.new(files[0],"SIGMAH",{:axis=>true})
          end
          gsigma = GPhys.new( Grid.new(zax), vsigma ) 
        end
        if axisnames.include?("t")
          xtimes.flatten!.sort!
          t = NArray.to_na(xtimes)
          vt = VArray.new( t ).rename("t")
          tax = Axis.new().set_pos(vt)
        end

        case ordering
        when "YX"
          grid = Grid.new(xax,yax,tax)
          grid.set_assoc_coords([glon,glat])
        when "YXS"
          grid = Grid.new(xax,yax,zax,tax)
          grid.set_assoc_coords([glon,glat,gsigma])
        when "YXW"
          grid = Grid.new(xax,yax,zax,tax)
          grid.set_assoc_coords([glon,glat,gsigma])
        else
          raise "ordering #{ordering} not impremented yet"
        end

        return GPhys.new(grid,vary)
      end
    end
  end
end

###############################################

if __FILE__ == $0
  puts "read MMOUT_DOMAIN2_??"
  p files = Dir.glob("MMOUT_DOMAIN2_??").sort!
  puts "create composite GPhys..."
  files = files.collect{|f| MM5IO.open(f)}
  p g = GPhys::MM5_IO::open(files,"U")
  puts
  p g = GPhys::MM5_IO::open(files,"W")
  puts
  p g = GPhys::MM5_IO::open(files,"T")
end
